@@ -1,5 +1,5 @@
 /* This file is part of GNU Rush.                  
-   Copyright (C) 2008-2011, 2014 Sergey Poznyakoff
+   Copyright (C) 2008-2019 Sergey Poznyakoff
 
    GNU Rush is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -46,7 +46,6 @@
 #include <regex.h>
 #include <c-ctype.h>
 #include <inttostr.h>
-#include "argmatch.h"
 
 #include <defines.h>
 #include <librush.h>
@@ -101,7 +100,9 @@ enum transform_node_type {
 	transform_map,
 	transform_delarg,
 	transform_setcmd,
-	transform_setarg
+	transform_setarg,
+	transform_setvar,
+	transform_unsetvar,
 };
 
 struct transform_node {
@@ -114,6 +115,7 @@ struct transform_node {
 		transform_t trans;
 		struct rush_map map;
 		int arg_end;
+		char *varname;
 	} v;
 };
 
@@ -181,7 +183,7 @@ struct rush_rule {
 	
 	/* Source location */
 	const char *file;                 /* Configuration file name */
-	size_t line;                      /* and line number. */
+	int line;                         /* and line number. */
 	
 	/* Match parameters */
 	struct test_node *test_head, *test_tail;
@@ -210,9 +212,16 @@ struct rush_rule {
 	struct rush_sockaddr post_sockaddr;
 };
 
+struct rush_backref {
+	char *subject;
+ 	regmatch_t *match;
+	size_t nmatch;
+	size_t maxmatch;
+};
+
 struct rush_request {
         char *cmdline;         /* Command line */
-        int argc;              /* Number of elements in argv */
+        size_t argc;           /* Number of elements in argv */
         char **argv;           /* Command line in parsed form */
 	int interactive;       /* Request for interactive shell */
 	char *prog;            /* Program file name, if different
@@ -226,6 +235,23 @@ struct rush_request {
 	enum rush_three_state acct;
 	const struct rush_sockaddr *post_sockaddr;
 	struct rush_i18n i18n;
+
+	/* Backreferences
+	   The backref field contains backreferences from the recent and
+	   penultimate regex matches. The backref_cur field indexes the
+	   recent backreferences. Remaining slot is used as a temporary
+	   storage during eventual next match. If that match succeeds, the
+	   value of backref_cur is updated to reflect the fact. The
+	   backref_count field keeps the actual number of backreferences
+	   used in the recent match. */
+	struct rush_backref backref[2]; 
+	int backref_cur;
+	size_t backref_count;
+
+	/* Temporary variable storage */
+	char **var_kv;
+	size_t var_count;
+	size_t var_max;
 };
 
 #define PROGFILE(req) ((req)->prog ? (req)->prog : (req)->argv[0])
@@ -237,56 +263,17 @@ extern int lint_option;
 extern unsigned sleep_time;
 extern struct rush_rule *rule_head, *rule_tail;
 extern unsigned debug_level;
+extern char *dump_option;
 extern int debug_option;
 extern struct passwd *rush_pw;
 
 #define __debug_p(x) ((x) <= debug_level)
 
-#define debug(lev,msg)					\
-	do {						\
-		if (__debug_p(lev))			\
-			logmsg(LOG_DEBUG, "%s", msg);	\
+#define debug(lev,fmt,...)					\
+	do {							\
+		if (__debug_p(lev))				\
+			logmsg(LOG_DEBUG, fmt, __VA_ARGS__);	\
 	} while(0)	
-#define debug1(lev,fmt,x1)				\
-	do {						\
-		if (__debug_p(lev))			\
-			logmsg(LOG_DEBUG, fmt, x1);	\
-	} while(0)
-#define debug2(lev,fmt,x1,x2)				\
-	do {						\
-		if (__debug_p(lev))			\
-			logmsg(LOG_DEBUG, fmt, x1, x2);	\
-	} while(0)
-#define debug3(lev,fmt,x1,x2,x3)		        \
-	do {						\
-		if (__debug_p(lev))			\
-			logmsg(LOG_DEBUG, fmt, x1, x2,	\
-			       x3);			\
-	} while(0)
-#define debug4(lev,fmt,x1,x2,x3,x4)		        \
-	do {						\
-		if (__debug_p(lev))			\
-			logmsg(LOG_DEBUG, fmt, x1, x2,	\
-			       x3, x4);			\
-	} while(0)
-#define debug5(lev,fmt,x1,x2,x3,x4,x5)		        \
-	do {						\
-		if (__debug_p(lev))			\
-			logmsg(LOG_DEBUG, fmt, x1, x2,	\
-			       x3, x4, x5);		\
-	} while(0)
-#define debug6(lev,fmt,x1,x2,x3,x4,x5,x6)	        \
-	do {						\
-		if (__debug_p(lev))			\
-			logmsg(LOG_DEBUG, fmt, x1, x2,	\
-			       x3, x4, x5, x6);		\
-	} while(0)
-#define debug7(lev,fmt,x1,x2,x3,x4,x5,x6,x7)	        \
-	do {						\
-		if (__debug_p(lev))			\
-			logmsg(LOG_DEBUG, fmt, x1, x2,	\
-			       x3, x4, x5, x6, x7);	\
-	} while(0)
 
 void die(enum error_type type, struct rush_i18n *i18n, const char *fmt, ...)
 	 RUSH_NORETURN RUSH_PRINTFLIKE(3,4);
@@ -332,5 +319,8 @@ int cfck_keyword(const char *name);
 /* map.c */
 char *map_string(struct rush_map *map, struct rush_request *req);
 char *rush_expand_string(const char *string, struct rush_request *req);
+
+/* dump.c */
+void dump_request(struct rush_request *req, FILE *fp);
 
 
